use core::panic;
use std::{fs, path::PathBuf};

use pest::Parser;
use pest_derive::Parser;

#[derive(Parser)]
#[grammar = "rpy.pest"]
struct RpyParser;

// Raw script tokens
#[derive(Debug, Clone)]
enum Token {
    Keyword(String),
    Str(String),
    Array(Vec<Token>),
    Boolean(bool),
    Number(f64),
}

impl Token {
    fn print(&self) -> &str {
        match &self {
            Keyword(keyword) => keyword,
            Str(_) => "String",
            Array(_) => "Array",
            Boolean(_) => "Boolean",
            Number(_) => "Number",
        }
    }
}

use Token::*;

// Indented command block
#[derive(Debug)]
struct CommandBlock {
    elements: Vec<BlockElement>,
    next: Option<usize>,
}

impl CommandBlock {
    fn next(&mut self) -> Option<&Command> {
        let mut next = match self.next {
            Some(next) => next,
            None => return None,
        };
        let mut result = None;
        let count = self.elements.len();
        for element in &mut self.elements[next..] {
            match element {
                BlockElement::Command(command) => {
                    result = Some(&*command);
                    next += 1;
                    break;
                },
                BlockElement::Block(block) => {
                    match block.next() {
                        Some(command) => {
                            result = Some(command);
                            break;
                        },
                        None => {
                            next += 1;
                        }
                    }
                }
            };
        }
        self.next = if count >= next {
            Some(next)
        } else {
            None
        };
        result
    }
}

impl Default for CommandBlock {
    fn default() -> Self {
        Self {
            elements: Vec::new(),
            next: Some(0),
        }
    }
}

#[derive(Debug)]
enum BlockElement {
    Command(Command),
    Block(CommandBlock),
}

type Script = CommandBlock;

// Parsed script commands
#[derive(Debug)]
#[allow(dead_code)]
enum Command {
    Say { name: Option<String>, text: String },
    Eat { food: String, politely: bool },
}

use Command::*;

impl Command {
    fn is_blocking(&self) -> bool {
        match self {
            Say { .. } => true,
            _ => false,
        }
    }
}

#[derive(Debug)]
pub enum Event {
    Say { name: Option<String>, text: String },
}

// ======
// Parser
// ======

type Pair<'a> = pest::iterators::Pair<'a, Rule>;

// Read file into commands
fn parse_file(file_path: &PathBuf) -> CommandBlock {
    let unparsed_file = fs::read_to_string(file_path).expect("cannot find file");
    parse(&unparsed_file)
}

fn parse(script: &str) -> CommandBlock {
    let file = RpyParser::parse(Rule::File, script)
        .expect("unsuccessful parse")
        .next()
        .unwrap();
    parse_block(file)
}

fn parse_block(block: Pair) -> CommandBlock {
    CommandBlock {
        elements: block.into_inner().filter_map(|pair| match pair.as_rule() {
            Rule::Line => Some(BlockElement::Command(parse_command(pair))),
            Rule::Block => Some(BlockElement::Block(parse_block(pair))),
            Rule::EOI => None, // end
            _ => unreachable!(),
        }).collect(),
        ..Default::default()
    }
}

fn parse_command(pair: Pair) -> Command {
    use Token::*;
    let line: Vec<Token> = pair
        .into_inner()
        .map(|pair| parse_token(pair))
        .collect();
    macro_rules! unknown {
        () => {
            panic!("Unknown command {}", describe_line(&line))
        };
    }
    match line.as_slice() {
        [Str(text)] => Say {
            name: None,
            text: text.to_owned(),
        },
        [Str(name), Str(text)] => Say {
            name: Some(name.to_owned()),
            text: text.to_owned(),
        },
        [Keyword(keyword), Str(food), tail @ ..] if keyword.eq("eat") => Eat {
            food: food.to_owned(),
            politely: match tail {
                [Boolean(politely)] => *politely,
                _ => unknown!(),
            },
        },
        _ => unknown!(),
    }
}

// Line description e.g. [String, Keyword, Array]
// Used in parse_command as feedback for invalid commands
fn describe_line(line: &[Token]) -> String {
    let mut description = "[".to_owned();
    let mut iter = line.iter();
    description.push_str(&format!("{}", iter.next().unwrap().print()));
    for token in iter {
        description.push_str(&format!(", {}", token.print()));
    }
    description.push_str("]");
    description
}

fn parse_token(pair: Pair) -> Token {
    let token = pair.as_rule();
    macro_rules! contents {
        () => {
            pair.into_inner().next().unwrap()
        };
    }
    match token {
        Rule::String => {
            let contents = contents!();
            Token::Str(match contents.as_rule() {
                Rule::SingleQuoteStringData => contents.as_str().replace("\\'", "'"),
                Rule::DoubleQuoteStringData => contents.as_str().replace("\\\"", "\""),
                _ => unreachable!(),
            })
        }
        Rule::Array => {
            let contents = contents!();
            let mut array = Vec::new();
            for token in contents.into_inner() {
                array.push(parse_token(token));
            }
            Token::Array(array)
        }
        Rule::Boolean => Token::Boolean(match pair.as_str() {
            "True" => true,
            "False" => false,
            _ => unreachable!(),
        }),
        Rule::Number => Token::Number(pair.as_str().parse().unwrap()),
        Rule::Keyword => Token::Keyword(pair.as_str().to_owned()),
        __ => unreachable!(),
    }
}

// =====
// State
// =====

pub struct State {
    script: Script,
}

impl State {
    pub fn from_file(file: PathBuf) -> State {
        State {
            script: parse_file(&file),
        }
    }

    pub fn next(&mut self) -> Option<Event> {
        while let Some(command) = self.next_command() {
            if command.is_blocking() {
                return Some(match command {
                    Say { name, text } => Event::Say {
                        name: name.clone(),
                        text: text.clone()
                    },
                    _ => unimplemented!(),
                });
            }
        }
        None
    }

    fn next_command(&mut self) -> Option<&Command> {
        self.script.next()
    }
}
